@@ -17,6 +17,13 @@
 #include <sys/ttydefaults.h>
 #endif
 
+#ifdef USE_TTY_GROUP
+# define TTY_MODE 0620
+#else
+# define TTY_MODE 0600
+#endif
+#define TTYGRPNAME      "tty"   /* name of group to own ttys */
+
 /* Some shorthands for control characters. */
 #define CTL(x)		((x) ^ 0100)	/* Assumes ASCII dialect */
 #define CR		CTL('M')	/* carriage return */
@@ -73,7 +80,7 @@ struct chardata {
 #define INIT_CHARDATA(ptr) do {              \
 		(ptr)->erase    = DEF_ERASE; \
 		(ptr)->kill     = DEF_KILL;  \
-		(ptr)->eol      = CTRL('r'); \
+		(ptr)->eol      = CR;        \
 	        (ptr)->parity   = 0;         \
 	        (ptr)->capslock = 0;         \
 	} while (0)
@@ -81,6 +88,7 @@ struct chardata {
 extern int get_terminal_dimension(int *cols, int *lines);
 extern int get_terminal_width(int default_width);
 extern int get_terminal_type(const char **type);
+extern int get_terminal_stdfd(void);
 extern int get_terminal_name(const char **path, const char **name,
 			     const char **number);
 
@@ -156,9 +164,10 @@ static inline void reset_virtual_console(struct termios *tp, int flags)
 	tp->c_oflag &= ~(OLCUC | OCRNL | ONOCR | ONLRET | OFILL | \
 			    NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
 	tp->c_lflag |=  (ISIG | ICANON | IEXTEN | ECHO|ECHOE|ECHOK|ECHOKE|ECHOCTL);
-	tp->c_lflag &= ~(ECHONL | NOFLSH | TOSTOP);
 #ifdef ECHOPRT
-	tp->c_lflag &= ~ECHOPRT;
+	tp->c_lflag &= ~(ECHONL|ECHOPRT | NOFLSH | TOSTOP);
+#else
+	tp->c_lflag &= ~(ECHONL| NOFLSH | TOSTOP);
 #endif
 
 	if ((flags & UL_TTY_KEEPCFLAGS) == 0) {
