@@ -31,6 +31,7 @@
 #include <sys/socket.h>
 #include <langinfo.h>
 #include <grp.h>
+#include <pwd.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <ifaddrs.h>
@@ -42,11 +43,14 @@
 #include "nls.h"
 #include "pathnames.h"
 #include "c.h"
+#include "cctype.h"
 #include "widechar.h"
 #include "ttyutils.h"
 #include "color-names.h"
 #include "env.h"
 
+#include "logindefs.h"
+
 #ifdef USE_PLYMOUTH_SUPPORT
 # include "plymouth-ctrl.h"
 #endif
@@ -55,6 +59,10 @@
 # include <sys/param.h>
 #endif
 
+#ifdef HAVE_GETTTYNAM
+# include <ttyent.h>
+#endif
+
 #if defined(__FreeBSD_kernel__)
 # include <pty.h>
 # ifdef HAVE_UTMP_H
@@ -147,8 +155,10 @@
 #endif
 
 /* Login prompt. */
-#define LOGIN		"login: "
-#define LOGIN_ARGV_MAX	16		/* Numbers of args for login */
+#define LOGIN_PROMPT		"login: "
+
+/* Numbers of args for login(1) */
+#define LOGIN_ARGV_MAX	16
 
 /*
  * agetty --reload
@@ -193,8 +203,8 @@ struct options {
 	char *chroot;			/* Chroot before the login */
 	char *login;			/* login program */
 	char *logopt;			/* options for login program */
-	char *tty;			/* name of tty */
-	char *vcline;			/* line of virtual console */
+	const char *tty;		/* name of tty */
+	const char *vcline;		/* line of virtual console */
 	char *term;			/* terminal type */
 	char *initstring;		/* modem init string */
 	char *issue;			/* alternative issue file or directory */
@@ -206,6 +216,7 @@ struct options {
 	int numspeed;			/* number of baud rates to try */
 	int clocal;			/* CLOCAL_MODE_* */
 	int kbmode;			/* Keyboard mode if virtual console */
+	int tty_is_stdin;		/* is the tty the standard input stream */
 	speed_t speeds[MAX_SPEED];	/* baud rates to be tried */
 };
 
@@ -322,7 +333,7 @@ static void init_special_char(char* arg, struct options *op);
 static void parse_args(int argc, char **argv, struct options *op);
 static void parse_speeds(struct options *op, char *arg);
 static void update_utmp(struct options *op);
-static void open_tty(char *tty, struct termios *tp, struct options *op);
+static void open_tty(const char *tty, struct termios *tp, struct options *op);
 static void termio_init(struct options *op, struct termios *tp);
 static void reset_vc(const struct options *op, struct termios *tp, int canon);
 static void auto_baud(struct termios *tp);
@@ -349,6 +360,8 @@ static void login_options_to_argv(char *argv[], int *argc, char *str, char *user
 static void reload_agettys(void);
 static void print_issue_file(struct issue *ie, struct options *op, struct termios *tp);
 static void eval_issue_file(struct issue *ie, struct options *op, struct termios *tp);
+static void show_issue(struct options *op);
+
 
 /* Fake hostname for ut_host specified on command line. */
 static char *fakehost;
@@ -395,13 +408,17 @@ int main(int argc, char **argv)
 	sigaction(SIGINT, &sa, &sa_int);
 
 #ifdef DEBUGGING
-	dbf = fopen(DEBUG_OUTPUT, "w");
-	for (int i = 1; i < argc; i++) {
-		if (i > 1)
-			debug(" ");
-		debug(argv[i]);
+	{
+		int i;
+
+		dbf = fopen(DEBUG_OUTPUT, "w");
+		for (i = 1; i < argc; i++) {
+			if (i > 1)
+				debug(" ");
+			debug(argv[i]);
+		}
+		debug("\n");
 	}
-	debug("\n");
 #endif				/* DEBUGGING */
 
 	/* Parse command-line arguments. */
@@ -440,7 +457,7 @@ int main(int argc, char **argv)
 	termio_init(&options, &termios);
 
 	/* Write the modem init string and DO NOT flush the buffers. */
-	if (serial_tty_option(&options, F_INITSTRING) &&
+	if (options.flags & F_INITSTRING &&
 	    options.initstring && *options.initstring != '\0') {
 		debug("writing init string\n");
 		write_all(STDOUT_FILENO, options.initstring,
@@ -489,12 +506,19 @@ int main(int argc, char **argv)
 	if (options.flags & F_NOPROMPT) {	/* --skip-login */
 		eval_issue_file(&issue, &options, &termios);
 		print_issue_file(&issue, &options, &termios);
+
 	} else {				/* regular (auto)login */
+		if ((options.flags & F_NOHOSTNAME) == 0 &&
+		    getlogindefs_bool("LOGIN_PLAIN_PROMPT", 0) == 1)
+			/* /etc/login.defs enbles --nohostname too */
+			options.flags |= F_NOHOSTNAME;
+
 		if (options.autolog) {
 			/* Autologin prompt */
 			eval_issue_file(&issue, &options, &termios);
 			do_prompt(&issue, &options, &termios);
-			printf(_("%s%s (automatic login)\n"), LOGIN, options.autolog);
+			printf(_("%s%s (automatic login)\n"), LOGIN_PROMPT,
+					options.autolog);
 		} else {
 			/* Read the login name. */
 			debug("reading login name\n");
@@ -542,8 +566,7 @@ int main(int argc, char **argv)
 		if (username) {
 			if (options.autolog)
 				login_argv[login_argc++] = "-f";
-			else
-				login_argv[login_argc++] = "--";
+			login_argv[login_argc++] = "--";
 			login_argv[login_argc++] = username;
 		}
 	}
@@ -585,7 +608,7 @@ static char *replace_u(char *str, char *username)
 		size_t sz;
 		char *tp, *old = entry;
 
-		if (memcmp(p, "\\u", 2)) {
+		if (memcmp(p, "\\u", 2) != 0) {
 			p++;
 			continue;	/* no \u */
 		}
@@ -601,16 +624,13 @@ static char *replace_u(char *str, char *username)
 		if (!tp)
 			log_err(_("failed to allocate memory: %m"));
 
-		if (p != str) {
+		if (p != str)
 			/* copy chars before \u */
-			memcpy(tp, str, p - str);
-			tp += p - str;
-		}
-		if (usz) {
+			tp = mempcpy(tp, str, p - str);
+		if (usz)
 			/* copy username */
-			memcpy(tp, username, usz);
-			tp += usz;
-		}
+			tp = mempcpy(tp, username, usz);
+
 		if (*(p + 2))
 			/* copy chars after \u + \0 */
 			memcpy(tp, p + 2, sz - (p - str) - 1);
@@ -705,6 +725,7 @@ static void output_version(void)
 static void parse_args(int argc, char **argv, struct options *op)
 {
 	int c;
+	int opt_show_issue = 0;
 
 	enum {
 		VERSION_OPTION = CHAR_MAX + 1,
@@ -716,6 +737,7 @@ static void parse_args(int argc, char **argv, struct options *op)
 		KILL_CHARS_OPTION,
 		RELOAD_OPTION,
 		LIST_SPEEDS_OPTION,
+		ISSUE_SHOW_OPTION,
 	};
 	const struct option longopts[] = {
 		{  "8bits",	     no_argument,	 NULL,  '8'  },
@@ -725,6 +747,7 @@ static void parse_args(int argc, char **argv, struct options *op)
 		{  "delay",	     required_argument,	 NULL,  'd'  },
 		{  "remote",         no_argument,        NULL,  'E'  },
 		{  "issue-file",     required_argument,  NULL,  'f'  },
+		{  "show-issue",     no_argument,        NULL,  ISSUE_SHOW_OPTION },
 		{  "flow-control",   no_argument,	 NULL,  'h'  },
 		{  "host",	     required_argument,  NULL,  'H'  },
 		{  "noissue",	     no_argument,	 NULL,  'i'  },
@@ -871,6 +894,9 @@ static void parse_args(int argc, char **argv, struct options *op)
 		case LIST_SPEEDS_OPTION:
 			list_speeds();
 			exit(EXIT_SUCCESS);
+		case ISSUE_SHOW_OPTION:
+			opt_show_issue = 1;
+			break;
 		case VERSION_OPTION:
 			output_version();
 			exit(EXIT_SUCCESS);
@@ -881,6 +907,11 @@ static void parse_args(int argc, char **argv, struct options *op)
 		}
 	}
 
+	if (opt_show_issue) {
+		show_issue(op);
+		exit(EXIT_SUCCESS);
+	}
+
 	debug("after getopt loop\n");
 
 	if (argc < optind + 1) {
@@ -908,6 +939,15 @@ static void parse_args(int argc, char **argv, struct options *op)
 		}
 	}
 
+	/* resolve the tty path in case it was provided as stdin */
+	if (strcmp(op->tty, "-") == 0) {
+		op->tty_is_stdin = 1;
+		int fd = get_terminal_name(NULL, &op->tty, NULL);
+		if (fd < 0) {
+			log_warn(_("could not get terminal name: %d"), fd);
+		}
+	}
+
 	/* On virtual console remember the line which is used for */
 	if (strncmp(op->tty, "tty", 3) == 0 &&
 	    strspn(op->tty + 3, "0123456789") == strlen(op->tty+3))
@@ -948,8 +988,8 @@ static void update_utmp(struct options *op)
 	time_t t;
 	pid_t pid = getpid();
 	pid_t sid = getsid(0);
-	char *vcline = op->vcline;
-	char *line   = op->tty;
+	const char *vcline = op->vcline;
+	const char *line = op->tty;
 	struct utmpx *utp;
 
 	/*
@@ -988,7 +1028,7 @@ static void update_utmp(struct options *op)
 			str2memcpy(ut.ut_id, vcline, sizeof(ut.ut_id));
 		else {
 			size_t len = strlen(line);
-			char * ptr;
+			const char * ptr;
 			if (len >= sizeof(ut.ut_id))
 				ptr = line + len - sizeof(ut.ut_id);
 			else
@@ -1018,7 +1058,7 @@ static void update_utmp(struct options *op)
 #endif				/* SYSV_STYLE */
 
 /* Set up tty as stdin, stdout & stderr. */
-static void open_tty(char *tty, struct termios *tp, struct options *op)
+static void open_tty(const char *tty, struct termios *tp, struct options *op)
 {
 	const pid_t pid = getpid();
 	int closed = 0;
@@ -1028,7 +1068,7 @@ static void open_tty(char *tty, struct termios *tp, struct options *op)
 
 	/* Set up new standard input, unless we are given an already opened port. */
 
-	if (strcmp(tty, "-") != 0) {
+	if (!op->tty_is_stdin) {
 		char buf[PATH_MAX+1];
 		struct group *gr = NULL;
 		struct stat st;
@@ -1156,6 +1196,18 @@ static void open_tty(char *tty, struct termios *tp, struct options *op)
 	if (tcgetattr(STDIN_FILENO, tp) < 0)
 		log_err(_("%s: failed to get terminal attributes: %m"), tty);
 
+#ifdef HAVE_GETTTYNAM
+	if (!op->term) {
+		struct ttyent *ent = getttynam(tty);
+		/* a bit nasty as it's never freed */
+		if (ent && ent->ty_type) {
+			op->term = strdup(ent->ty_type);
+			if (!op->term)
+				log_err(_("failed to allocate memory: %m"));
+		}
+	}
+#endif
+
 #if defined (__s390__) || defined (__s390x__)
 	if (!op->term) {
 	        /*
@@ -1284,9 +1336,17 @@ static void termio_init(struct options *op, struct termios *tp)
 		ispeed = cfgetispeed(tp);
 		ospeed = cfgetospeed(tp);
 
-		if (!ispeed) ispeed = TTYDEF_SPEED;
-		if (!ospeed) ospeed = TTYDEF_SPEED;
-
+		/* Save also the original speed to array of the speeds to make
+		 * it possible to return the original after unexpected BREAKs.
+		 */
+		if (op->numspeed)
+			op->speeds[op->numspeed++] = ispeed ? ispeed :
+						     ospeed ? ospeed :
+						     TTYDEF_SPEED;
+		if (!ispeed)
+			ispeed = TTYDEF_SPEED;
+		if (!ospeed)
+			ospeed = TTYDEF_SPEED;
 	} else {
 		ospeed = ispeed = op->speeds[FIRST_SPEED];
 	}
@@ -1298,7 +1358,7 @@ static void termio_init(struct options *op, struct termios *tp)
 	 * later on.
 	 */
 
-	/* The defaul is set c_iflag in termio_final() according to chardata.
+	/* The default is set c_iflag in termio_final() according to chardata.
 	 * Unfortunately, the chardata are not set according to the serial line
 	 * if --autolog is enabled. In this case we do not read from the line
 	 * at all. The best what we can do in this case is to keep c_iflag
@@ -1446,7 +1506,7 @@ static void auto_baud(struct termios *tp)
 	if ((nread = read(STDIN_FILENO, buf, sizeof(buf) - 1)) > 0) {
 		buf[nread] = '\0';
 		for (bp = buf; bp < buf + nread; bp++)
-			if (isascii(*bp) && isdigit(*bp)) {
+			if (c_isascii(*bp) && isdigit(*bp)) {
 				if ((speed = bcode(bp))) {
 					cfsetispeed(tp, speed);
 					cfsetospeed(tp, speed);
@@ -1706,7 +1766,9 @@ static int wait_for_term_input(int fd)
 		if (FD_ISSET(fd, &rfds)) {
 			return 1;
 
-		} else if (netlink_fd >= 0 && FD_ISSET(netlink_fd, &rfds)) {
+		}
+
+		if (netlink_fd >= 0 && FD_ISSET(netlink_fd, &rfds)) {
 			if (!process_netlink())
 				continue;
 
@@ -1735,33 +1797,61 @@ static int issuedir_filter(const struct dirent *d)
 
 	namesz = strlen(d->d_name);
 	if (!namesz || namesz < ISSUEDIR_EXTSIZ + 1 ||
-	    strcmp(d->d_name + (namesz - ISSUEDIR_EXTSIZ), ISSUEDIR_EXT))
+	    strcmp(d->d_name + (namesz - ISSUEDIR_EXTSIZ), ISSUEDIR_EXT) != 0)
 		return 0;
 
 	/* Accept this */
 	return 1;
 }
 
-static FILE *issuedir_next_file(int dd, struct dirent **namelist, int nfiles, int *n)
+
+static int issuefile_read_stream(struct issue *ie, FILE *f, struct options *op, struct termios *tp);
+
+/* returns: 0 on success, 1 cannot open, <0 on error
+ */
+static int issuedir_read(struct issue *ie, const char *dirname,
+			 struct options *op, struct termios *tp)
 {
-	while (*n < nfiles) {
-		struct dirent *d = namelist[*n];
-		struct stat st;
+        int dd, nfiles, i;
+        struct dirent **namelist = NULL;
+
+	dd = open(dirname, O_RDONLY|O_CLOEXEC|O_DIRECTORY);
+	if (dd < 0)
+		return 1;
+
+	nfiles = scandirat(dd, ".", &namelist, issuedir_filter, versionsort);
+	if (nfiles <= 0)
+		goto done;
+
+	ie->do_tcsetattr = 1;
+
+	for (i = 0; i < nfiles; i++) {
+		struct dirent *d = namelist[i];
 		FILE *f;
 
-		(*n)++;
-
-		if (fstatat(dd, d->d_name, &st, 0) ||
-		    !S_ISREG(st.st_mode))
-			continue;
-
 		f = fopen_at(dd, d->d_name, O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
-		if (f)
-			return f;
+		if (f) {
+			issuefile_read_stream(ie, f, op, tp);
+			fclose(f);
+		}
 	}
-	return NULL;
+
+	for (i = 0; i < nfiles; i++)
+		free(namelist[i]);
+	free(namelist);
+done:
+	close(dd);
+	return 0;
 }
 
+#else /* !ISSUEDIR_SUPPORT */
+static int issuedir_read(struct issue *ie __attribute__((__unused__)),
+			const char *dirname __attribute__((__unused__)),
+			struct options *op __attribute__((__unused__)),
+			struct termios *tp __attribute__((__unused__)))
+{
+	return 1;
+}
 #endif /* ISSUEDIR_SUPPORT */
 
 #ifndef ISSUE_SUPPORT
@@ -1780,8 +1870,55 @@ static void eval_issue_file(struct issue *ie __attribute__((__unused__)),
 			    struct termios *tp __attribute__((__unused__)))
 {
 }
+
+static void show_issue(struct options *op __attribute__((__unused__)))
+{
+}
+
 #else /* ISSUE_SUPPORT */
 
+static int issuefile_read_stream(
+		struct issue *ie, FILE *f,
+		struct options *op, struct termios *tp)
+{
+	struct stat st;
+	int c;
+
+	if (fstat(fileno(f), &st) || !S_ISREG(st.st_mode))
+		return 1;
+
+	if (!ie->output) {
+		free(ie->mem);
+		ie->mem_sz = 0;
+		ie->mem = NULL;
+		ie->output = open_memstream(&ie->mem, &ie->mem_sz);
+	}
+
+	while ((c = getc(f)) != EOF) {
+		if (c == '\\')
+			output_special_char(ie, getc(f), op, tp, f);
+		else
+			putc(c, ie->output);
+	}
+
+	return 0;
+}
+
+static int issuefile_read(
+		struct issue *ie, const char *filename,
+		struct options *op, struct termios *tp)
+{
+	FILE *f = fopen(filename, "r" UL_CLOEXECSTR);
+	int rc = 1;
+
+	if (f) {
+		rc = issuefile_read_stream(ie, f, op, tp);
+		fclose(f);
+	}
+	return rc;
+}
+
+
 #ifdef AGETTY_RELOAD
 static int issue_is_changed(struct issue *ie)
 {
@@ -1842,100 +1979,98 @@ static void eval_issue_file(struct issue *ie,
 			    struct options *op,
 			    struct termios *tp)
 {
-	const char *filename, *dirname = NULL;
-	FILE *f = NULL;
-#ifdef ISSUEDIR_SUPPORT
-	int dd = -1, nfiles = 0, i;
-	struct dirent **namelist = NULL;
-#endif
+	int has_file = 0;
+
 #ifdef AGETTY_RELOAD
 	netlink_groups = 0;
 #endif
-
 	if (!(op->flags & F_ISSUE))
-		return;
-
+		goto done;
 	/*
-	 * The custom issue file or directory specified by: agetty -f <path>.
+	 * The custom issue file or directory list specified by:
+	 *   agetty --isue-file <path[:path]...>
 	 * Note that nothing is printed if the file/dir does not exist.
 	 */
-	filename = op->issue;
-	if (filename) {
-		struct stat st;
+	if (op->issue) {
+		char *list = strdup(op->issue);
+		char *file;
 
-		if (stat(filename, &st) < 0)
-			return;
-		if (S_ISDIR(st.st_mode)) {
-			dirname = filename;
-			filename = NULL;
+		if (!list)
+			log_err(_("failed to allocate memory: %m"));
+
+		for (file = strtok(list, ":"); file; file = strtok(NULL, ":")) {
+			struct stat st;
+
+			if (stat(file, &st) < 0)
+				continue;
+			if (S_ISDIR(st.st_mode))
+				issuedir_read(ie, file, op, tp);
+			else
+				issuefile_read(ie, file, op, tp);
 		}
-	} else {
-		/* The default /etc/issue and optional /etc/issue.d directory
-		 * as extension to the file. The /etc/issue.d directory is
-		 * ignored if there is no /etc/issue file. The file may be
-		 * empty or symlink.
-		 */
-		if (access(_PATH_ISSUE, F_OK|R_OK) != 0)
-			return;
-		filename  = _PATH_ISSUE;
-		dirname = _PATH_ISSUEDIR;
+		free(list);
+		goto done;
 	}
 
-	ie->output = open_memstream(&ie->mem, &ie->mem_sz);
-#ifdef ISSUEDIR_SUPPORT
-	if (dirname) {
-		dd = open(dirname, O_RDONLY|O_CLOEXEC|O_DIRECTORY);
-		if (dd >= 0)
-			nfiles = scandirat(dd, ".", &namelist, issuedir_filter, versionsort);
-		if (nfiles <= 0)
-			dirname = NULL;
-	}
-	i = 0;
-#endif
-	if (filename)
-		f = fopen(filename, "r");
-
-	if (f || dirname) {
-		int c;
-
-		ie->do_tcsetattr = 1;
-
-		do {
-#ifdef ISSUEDIR_SUPPORT
-			if (!f && i < nfiles)
-				f = issuedir_next_file(dd, namelist, nfiles, &i);
-#endif
-			if (!f)
-				break;
-			while ((c = getc(f)) != EOF) {
-				if (c == '\\')
-					output_special_char(ie, getc(f), op, tp, f);
-				else
-					putc(c, ie->output);
-			}
-			fclose(f);
-			f = NULL;
-		} while (dirname);
-
-		fflush(stdout);
-
-		if ((op->flags & F_VCONSOLE) == 0)
-			ie->do_tcrestore = 1;
+	/* The default /etc/issue and optional /etc/issue.d directory as
+	 * extension to the file. The /etc/issue.d directory is ignored if
+	 * there is no /etc/issue file. The file may be empty or symlink.
+	 */
+	if (access(_PATH_ISSUE, F_OK|R_OK) == 0) {
+		issuefile_read(ie, _PATH_ISSUE, op, tp);
+		issuedir_read(ie, _PATH_ISSUEDIR, op, tp);
+		goto done;
 	}
 
-#ifdef ISSUEDIR_SUPPORT
-	for (i = 0; i < nfiles; i++)
-		free(namelist[i]);
-	free(namelist);
-	if (dd >= 0)
-		close(dd);
-#endif
+	/* Fallback @runstatedir (usually /run) -- the file is not required to
+	 * read the dir.
+	 */
+	if (issuefile_read(ie, _PATH_RUNSTATEDIR "/" _PATH_ISSUE_FILENAME, op, tp) == 0)
+		has_file++;
+	if (issuedir_read(ie, _PATH_RUNSTATEDIR "/" _PATH_ISSUE_DIRNAME, op, tp) == 0)
+		has_file++;
+	if (has_file)
+		goto done;
+
+	/* Fallback @sysconfstaticdir (usually /usr/lib) -- the file is not
+	 * required to read the dir
+	 */
+	issuefile_read(ie, _PATH_SYSCONFSTATICDIR "/" _PATH_ISSUE_FILENAME, op, tp); 
+	issuedir_read(ie, _PATH_SYSCONFSTATICDIR "/" _PATH_ISSUE_DIRNAME, op, tp);
+
+done:
+
 #ifdef AGETTY_RELOAD
 	if (netlink_groups != 0)
 		open_netlink();
 #endif
-	fclose(ie->output);
+	if (ie->output) {
+		fclose(ie->output);
+		ie->output = NULL;
+	}
 }
+
+/* This is --show-issue backend, executed by normal user on the current
+ * terminal.
+ */
+static void show_issue(struct options *op)
+{
+	struct issue ie = { .output = NULL };
+	struct termios tp;
+
+	memset(&tp, 0, sizeof(struct termios));
+	if (tcgetattr(STDIN_FILENO, &tp) < 0)
+		err(EXIT_FAILURE, _("failed to get terminal attributes: %m"));
+
+	eval_issue_file(&ie, op, &tp);
+
+	if (ie.mem_sz)
+		write_all(STDOUT_FILENO, ie.mem, ie.mem_sz);
+	if (ie.output)
+		fclose(ie.output);
+	free(ie.mem);
+}
+
 #endif /* ISSUE_SUPPORT */
 
 /* Show login prompt, optionally preceded by /etc/issue contents. */
@@ -1953,7 +2088,8 @@ again:
 		if (!wait_for_term_input(STDIN_FILENO)) {
 			eval_issue_file(ie, op, tp);
 			if (issue_is_changed(ie)) {
-				if (op->flags & F_VCONSOLE)
+				if ((op->flags & F_VCONSOLE)
+				    && (op->flags & F_NOCLEAR) == 0)
 					termio_clear(STDOUT_FILENO);
 				goto again;
 			}
@@ -2023,7 +2159,8 @@ again:
 	}
 	if (!op->autolog) {
 		/* Always show login prompt. */
-		write_all(STDOUT_FILENO, LOGIN, sizeof(LOGIN) - 1);
+		write_all(STDOUT_FILENO, LOGIN_PROMPT,
+				sizeof(LOGIN_PROMPT) - 1);
 	}
 }
 
@@ -2073,7 +2210,7 @@ static char *get_logname(struct issue *ie, struct options *op, struct termios *t
 		sleep(1);
 	tcflush(STDIN_FILENO, TCIFLUSH);
 
-	eightbit = (op->flags & F_EIGHTBITS);
+	eightbit = (op->flags & (F_EIGHTBITS|F_UTF8));
 	bp = logname;
 	*bp = '\0';
 
@@ -2094,7 +2231,8 @@ static char *get_logname(struct issue *ie, struct options *op, struct termios *t
 			if (!issue_is_changed(ie))
 				goto no_reload;
 			tcflush(STDIN_FILENO, TCIFLUSH);
-			if (op->flags & F_VCONSOLE)
+			if ((op->flags & F_VCONSOLE)
+			    && (op->flags & F_NOCLEAR) == 0)
 				termio_clear(STDOUT_FILENO);
 			bp = logname;
 			*bp = '\0';
@@ -2180,6 +2318,11 @@ static char *get_logname(struct issue *ie, struct options *op, struct termios *t
 				break;
 			case CTL('U'):
 				cp->kill = ascval;		/* set kill character */
+				/* fallthrough */
+			case CTL('C'):
+				if (key == CTL('C') && !(op->flags & F_VCONSOLE))
+					/* Ignore CTRL+C on serial line */
+					break;
 				while (bp > logname) {
 					if ((tp->c_lflag & ECHO) == 0)
 						write_all(1, erase[cp->parity], 3);
@@ -2189,8 +2332,6 @@ static char *get_logname(struct issue *ie, struct options *op, struct termios *t
 			case CTL('D'):
 				exit(EXIT_SUCCESS);
 			default:
-				if (!isascii(ascval) || !isprint(ascval))
-					break;
 				if ((size_t)(bp - logname) >= sizeof(logname) - 1)
 					log_err(_("%s: input overrun"), op->tty);
 				if ((tp->c_lflag & ECHO) == 0)
@@ -2336,7 +2477,14 @@ static int caps_lock(char *s)
 static speed_t bcode(char *s)
 {
 	const struct Speedtab *sp;
-	long speed = atol(s);
+	char *end = NULL;
+	long speed;
+
+	errno = 0;
+	speed = strtol(s, &end, 10);
+
+	if (errno || !end || end == s)
+		return 0;
 
 	for (sp = speedtab; sp->speed; sp++)
 		if (sp->speed == speed)
@@ -2360,7 +2508,8 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_(" -a, --autologin <user>     login the specified user automatically\n"), out);
 	fputs(_(" -c, --noreset              do not reset control mode\n"), out);
 	fputs(_(" -E, --remote               use -r <hostname> for login(1)\n"), out);
-	fputs(_(" -f, --issue-file <file>    display issue file\n"), out);
+	fputs(_(" -f, --issue-file <list>    display issue files or directories\n"), out);
+	fputs(_("     --show-issue           display issue file and exit\n"), out);
 	fputs(_(" -h, --flow-control         enable hardware flow control\n"), out);
 	fputs(_(" -H, --host <hostname>      specify login host\n"), out);
 	fputs(_(" -i, --noissue              do not display issue file\n"), out);
@@ -2409,46 +2558,40 @@ static void list_speeds(void)
  * Will be used by log_err() and log_warn() therefore
  * it takes a format as well as va_list.
  */
-#define	str2cpy(b,s1,s2)	strcat(strcpy(b,s1),s2)
-
-static void dolog(int priority, const char *fmt, va_list ap)
-{
-#ifndef	USE_SYSLOG
-	int fd;
+static void dolog(int priority
+#ifndef USE_SYSLOG
+		  __attribute__((__unused__))
 #endif
-	char buf[BUFSIZ];
-	char *bp;
-
+		  , const char *fmt, va_list ap)
+{
+#ifdef USE_SYSLOG
 	/*
 	 * If the diagnostic is reported via syslog(3), the process name is
 	 * automatically prepended to the message. If we write directly to
 	 * /dev/console, we must prepend the process name ourselves.
 	 */
-#ifdef USE_SYSLOG
-	buf[0] = '\0';
-	bp = buf;
-#else
-	str2cpy(buf, program_invocation_short_name, ": ");
-	bp = buf + strlen(buf);
-#endif				/* USE_SYSLOG */
-	vsnprintf(bp, sizeof(buf)-strlen(buf), fmt, ap);
-
-	/*
-	 * Write the diagnostic directly to /dev/console if we do not use the
-	 * syslog(3) facility.
-	 */
-#ifdef	USE_SYSLOG
 	openlog(program_invocation_short_name, LOG_PID, LOG_AUTHPRIV);
-	syslog(priority, "%s", buf);
+	vsyslog(priority, fmt, ap);
 	closelog();
 #else
+	/*
+	 * Write the diagnostic directly to /dev/console if we do not use
+	 * the syslog(3) facility.
+	 */
+	char buf[BUFSIZ];
+	char new_fmt[BUFSIZ];
+	int fd;
+
+	snprintf(new_fmt, sizeof(new_fmt), "%s: %s\r\n",
+		 program_invocation_short_name, fmt);
 	/* Terminate with CR-LF since the console mode is unknown. */
-	strcat(bp, "\r\n");
+	vsnprintf(buf, sizeof(buf), new_fmt, ap);
+
 	if ((fd = open("/dev/console", 1)) >= 0) {
 		write_all(fd, buf, strlen(buf));
 		close(fd);
 	}
-#endif				/* USE_SYSLOG */
+#endif	/* USE_SYSLOG */
 }
 
 static void exit_slowly(int code)
@@ -2607,9 +2750,12 @@ static void output_special_char(struct issue *ie,
 		char escname[UL_COLORNAME_MAXSZ];
 
 		if (get_escape_argument(fp, escname, sizeof(escname))) {
-			const char *esc = color_sequence_from_colorname(escname);
-			if (esc)
+			char *esc = color_get_sequence(escname);
+
+			if (esc) {
 				fputs(esc, ie->output);
+				free(esc);
+			}
 		} else
 			fputs("\033", ie->output);
 		break;
@@ -2668,24 +2814,21 @@ static void output_special_char(struct issue *ie,
 	case 't':
 	{
 		time_t now;
-		struct tm *tm;
+		struct tm tm;
 
 		time(&now);
-		tm = localtime(&now);
-
-		if (!tm)
-			break;
+		localtime_r(&now, &tm);
 
 		if (c == 'd') /* ISO 8601 */
-			fprintf(ie->output, "%s %s %d  %d",
-				      nl_langinfo(ABDAY_1 + tm->tm_wday),
-				      nl_langinfo(ABMON_1 + tm->tm_mon),
-				      tm->tm_mday,
-				      tm->tm_year < 70 ? tm->tm_year + 2000 :
-				      tm->tm_year + 1900);
+			fprintf(ie->output, "%s %s %2d  %d",
+				      nl_langinfo(ABDAY_1 + tm.tm_wday),
+				      nl_langinfo(ABMON_1 + tm.tm_mon),
+				      tm.tm_mday,
+				      tm.tm_year < 70 ? tm.tm_year + 2000 :
+				      tm.tm_year + 1900);
 		else
 			fprintf(ie->output, "%02d:%02d:%02d",
-				      tm->tm_hour, tm->tm_min, tm->tm_sec);
+				      tm.tm_hour, tm.tm_min, tm.tm_sec);
 		break;
 	}
 	case 'l':
@@ -2845,8 +2988,7 @@ static ssize_t append(char *dest, size_t len, const char  *sep, const char *src)
 
 	p = dest + dsz;
 	if (ssz) {
-		memcpy(p, sep, ssz);
-		p += ssz;
+		p = mempcpy(p, sep, ssz);
 	}
 	memcpy(p, src, sz);
 	*(p + sz) = '\0';
